"""
Standard test data for regridding benchmark.
"""

import numpy as np


def wave_smooth(lon, lat):
    r"""
    Spherical harmonic with low frequency.

    Parameters
    ----------
    lon, lat : 2D numpy array or xarray DataArray
         Longitute/Latitude of cell centers

    Returns
    -------
    f : 2D numpy array or xarray DataArray depending on input
        2D wave field

    Notes
    -----
    Equation from [1]_ [2]_:

    .. math:: Y_2^2 = 2 + \cos^2(\\theta) \cos(2 \phi)

    References
    ----------
    .. [1] Jones, P. W. (1999). First-and second-order conservative remapping
       schemes for grids in spherical coordinates. Monthly Weather Review,
       127(9), 2204-2210.

    .. [2] Ullrich, P. A., Lauritzen, P. H., & Jablonowski, C. (2009).
       Geometrically exact conservative remapping (GECoRe): regular
       latitude–longitude and cubed-sphere grids. Monthly Weather Review,
       137(6), 1721-1741.
    """
    # degree to radius, make a copy
    lat = lat / 180.0 * np.pi
    lon = lon / 180.0 * np.pi

    f = 2 + np.cos(lat) ** 2 * np.cos(2 * lon)
    return f



import cftime
import numpy as np
import pytest
import xarray as xr

import xesmf

mom6like = xr.Dataset(
    data_vars=dict(
        tos=(['time', 'yh', 'xh'], np.random.rand(2, 180, 360)),
    ),
    coords=dict(
        xq=xr.DataArray(
            np.arange(-300, 60 + 1),
            dims=['xq'],
            attrs={
                'long_name': 'q point nominal longitude',
                'units': 'degrees_east',
                'cartesian_axis': 'X',
            },
        ),
        yq=xr.DataArray(
            np.arange(-90, 90 + 1),
            dims=['yq'],
            attrs={
                'long_name': 'q point nominal latitude',
                'units': 'degrees_north',
                'cartesian_axis': 'Y',
            },
        ),
        xh=xr.DataArray(
            0.5 + np.arange(-300, 60),
            dims=['xh'],
            attrs={
                'long_name': 'h point nominal longitude',
                'units': 'degrees_east',
                'cartesian_axis': 'X',
            },
        ),
        yh=xr.DataArray(
            0.5 + np.arange(-90, 90),
            dims=['yh'],
            attrs={
                'long_name': 'h point nominal latitude',
                'units': 'degrees_north',
                'cartesian_axis': 'Y',
            },
        ),
        time=xr.DataArray(
            [
                cftime.DatetimeNoLeap(2007, 1, 16, 12, 0, 0, 0),
                cftime.DatetimeNoLeap(2007, 2, 15, 0, 0, 0, 0),
            ],
            dims=['time'],
        ),
        reference_time=cftime.DatetimeNoLeap(1901, 1, 1, 0, 0, 0, 0),
    ),
    attrs=dict(description='Synthetic MOM6 data'),
)


def test_mom6like_to_5x5():
    """regression test for MOM6 grid"""

    grid_5x5 = xr.Dataset()
    grid_5x5['lon'] = xr.DataArray(data=0.5 + np.arange(0, 360, 5), dims=('x'))
    grid_5x5['lat'] = xr.DataArray(data=0.5 - 90 + np.arange(0, 180, 5), dims=('y'))

    # multiple definition for lon/lat results in failure to determine
    # which coordinate set to use.
    with pytest.raises(ValueError):
        regrid_to_5x5 = xesmf.Regridder(mom6like, grid_5x5, 'bilinear', periodic=True)

    regrid_to_5x5 = xesmf.Regridder(
        mom6like.rename({'xh': 'lon', 'yh': 'lat'}), grid_5x5, 'bilinear', periodic=True
    )

    with pytest.warns(UserWarning, match=r"Using dimensions \('yh', 'xh'\)"):
        tos_regridded = regrid_to_5x5(mom6like['tos'])
    assert tos_regridded.shape == ((2, 36, 72))


@pytest.mark.testcases
def test_mom6_to_1x1():
    """check regridding of MOM6 to regular lon/lat"""

    dataurl = 'http://35.188.34.63:8080/thredds/dodsC/OM4p5/'
    mom6 = xr.open_dataset(
        f'{dataurl}/ocean_monthly_z.200301-200712.nc4',
        chunks={'time': 1, 'z_l': 1},
        drop_variables=['average_DT', 'average_T1', 'average_T2'],
        decode_times=False,
        engine='pydap',
    )

    grid_1x1 = xr.Dataset()
    grid_1x1['lon'] = xr.DataArray(data=0.5 + np.arange(360), dims=('x'))
    grid_1x1['lat'] = xr.DataArray(data=0.5 - 90 + np.arange(180), dims=('y'))

    regrid_to_1x1 = xesmf.Regridder(
        mom6.rename({'geolon': 'lon', 'geolat': 'lat'}), grid_1x1, 'bilinear', periodic=True
    )

    thetao_1x1 = regrid_to_1x1(mom6['thetao'])
    assert np.allclose(thetao_1x1.isel(time=0, z_l=0, x=200, y=100).values, 27.15691922)



import os
import warnings

import cf_xarray as cfxr
import dask
import numpy as np
import pytest
import xarray as xr
from dask.array.core import PerformanceWarning
from numpy.testing import assert_allclose, assert_almost_equal, assert_equal
from shapely import segmentize
from shapely.geometry import MultiPolygon, Polygon

import xesmf as xe
from xesmf.frontend import as_2d_mesh

dask_schedulers = ['threaded_scheduler', 'processes_scheduler', 'distributed_scheduler']
pytestmark = pytest.mark.filterwarnings('ignore:Input array is not C_CONTIGUOUS')

# same test data as test_backend.py, but here we can use xarray DataSet
ds_in = xe.util.grid_global(20, 12)
ds_in.lat.attrs['standard_name'] = 'latitude'
ds_in.lon.attrs['standard_name'] = 'longitude'
ds_out = xe.util.grid_global(15, 9)

horiz_shape_in = ds_in['lon'].shape
horiz_shape_out = ds_out['lon'].shape

ds_in['data'] = xe.data.wave_smooth(ds_in['lon'], ds_in['lat'])
ds_out['data_ref'] = xe.data.wave_smooth(ds_out['lon'], ds_out['lat'])

# 4D data to test broadcasting, increasing linearly with time and lev
ds_in.coords['time'] = np.arange(7) + 1
ds_in.coords['lev'] = np.arange(11) + 1
ds_in['data4D'] = ds_in['time'] * ds_in['lev'] * ds_in['data']
ds_in['data4D_f4'] = ds_in['data4D'].astype('f4')
ds_out['data4D_ref'] = ds_in['time'] * ds_in['lev'] * ds_out['data_ref']

# use non-divisible chunk size to catch edge cases
ds_in_chunked = ds_in.chunk({'time': 3, 'lev': 2})
ds_spatial_chunked = ds_in.chunk({'time': 3, 'lev': 2, 'y': 5, 'x': 9})
ds_out_chunked = ds_out.chunk({'y': 5, 'x': 5})

ds_locs = xr.Dataset()
ds_locs['lat'] = xr.DataArray(data=[-20, -10, 0, 10], dims=('locations',))
ds_locs['lon'] = xr.DataArray(data=[0, 5, 10, 15], dims=('locations',))
ds_locs_chunked = ds_locs.chunk(2)


# For polygon handling and spatial average
ds_savg = xr.Dataset(
    coords={
        'lat': (('lat',), [0.5, 1.5]),
        'lon': (('lon',), [0.5, 1.5, 2.5]),
        'lat_b': (('lat_b',), [0, 1, 2]),
        'lon_b': (('lon_b',), [0, 1, 2, 3]),
    },
    data_vars={'abc': (('lon', 'lat'), [[1.0, 2.0], [3.0, 4.0], [2.0, 4.0]])},
)
polys_raw = [
    Polygon([[0.5, 0.5], [0.5, 1.5], [1.5, 0.5]]),  # Simple triangle polygon
    MultiPolygon(
        [
            Polygon([[0.25, 1.25], [0.25, 1.75], [0.75, 1.75], [0.75, 1.25]]),
            Polygon([[1.25, 1.25], [1.25, 1.75], [1.75, 1.75], [1.75, 1.25]]),
        ]
    ),  # Multipolygon on 2 and 4
    Polygon(
        [[0, 0], [0, 1], [2, 1], [2, 0]],
        holes=[[[0.5, 0.25], [0.5, 0.75], [1.0, 0.75], [1.0, 0.25]]],
    ),  # Simple polygon covering 1 and 3 with hole over 1
    Polygon([[1, 1], [1, 3], [3, 3], [3, 1]]),  # Polygon partially outside, covering a part of 4
    Polygon([[3, 3], [3, 4], [4, 4], [4, 3]]),  # Polygon totally outside
    Polygon(
        [
            [0, 0],
            [0.5, 0.5],
            [0, 1],
            [0.5, 1.5],
            [0, 2],
            [2, 2],
            [1.5, 1.5],
            [2, 1],
            [1.5, 0.5],
            [2, 0],
        ]
    ),  # Long multifaceted polygon
    [
        Polygon([[0.5, 0.5], [0.5, 1.5], [1.5, 0.5]]),
        MultiPolygon(
            [
                Polygon([[0.25, 1.25], [0.25, 1.75], [0.75, 1.75], [0.75, 1.25]]),
                Polygon([[1, 1], [1, 2], [2, 2], [2, 1]]),
            ]
        ),
    ],  # Combination of Polygon and MultiPolygon with two different areas
]


def _segmentize(p):
    if isinstance(p, list):
        return list(map(_segmentize, p))
    return segmentize(p, 1)


polys = list(map(_segmentize, polys_raw))
exps_polys = [1.75, 3, 2.1429, 4, 0, 2.5, [1.75, 3.6]]


def test_as_2d_mesh():
    # 2D grid should not change
    lon2d = ds_in['lon'].values
    lat2d = ds_in['lat'].values
    assert_equal((lon2d, lat2d), as_2d_mesh(lon2d, lat2d))

    # 1D grid should become 2D
    lon1d = lon2d[0, :]
    lat1d = lat2d[:, 0]
    assert_equal((lon2d, lat2d), as_2d_mesh(lon1d, lat1d))

    # mix of 1D and 2D should fail
    with pytest.raises(ValueError):
        as_2d_mesh(lon1d, lat2d)


# 'patch' is too slow to test
methods_list = ['bilinear', 'conservative', 'nearest_s2d', 'nearest_d2s']


@pytest.mark.parametrize(
    'locstream_in,locstream_out,method,unmapped_to_nan',
    [
        (False, False, 'conservative', False),
        (False, False, 'bilinear', False),
        (False, True, 'bilinear', False),
        (False, False, 'nearest_s2d', False),
        (False, True, 'nearest_s2d', False),
        (True, False, 'nearest_s2d', False),
        (True, True, 'nearest_s2d', False),
        (False, False, 'nearest_d2s', False),
        (False, True, 'nearest_d2s', False),
        (True, False, 'nearest_d2s', False),
        (True, True, 'nearest_d2s', False),
        (False, False, 'conservative', True),
        (False, False, 'bilinear', True),
        (False, True, 'bilinear', True),
        (False, False, 'nearest_s2d', True),
        (False, True, 'nearest_s2d', True),
        (True, False, 'nearest_s2d', True),
        (True, True, 'nearest_s2d', True),
        (False, False, 'nearest_d2s', True),
        (False, True, 'nearest_d2s', True),
        (True, False, 'nearest_d2s', True),
        (True, True, 'nearest_d2s', True),
    ],
)
def test_build_regridder(method, locstream_in, locstream_out, unmapped_to_nan):
    din = ds_locs if locstream_in else ds_in
    dout = ds_locs if locstream_out else ds_out

    regridder = xe.Regridder(
        din, dout, method, locstream_in=locstream_in, locstream_out=locstream_out
    )

    # check screen output
    assert repr(regridder) == str(regridder)
    assert 'xESMF Regridder' in str(regridder)
    assert method in str(regridder)


def test_existing_weights():
    # the first run
    method = 'bilinear'
    regridder = xe.Regridder(ds_in, ds_out, method)
    fn = regridder.to_netcdf()

    # make sure we can reuse weights
    assert os.path.exists(fn)
    regridder_reuse = xe.Regridder(ds_in, ds_out, method, weights=fn)
    assert regridder_reuse.weights.shape == regridder.weights.shape

    # this should also work with reuse_weights=True
    regridder_reuse = xe.Regridder(ds_in, ds_out, method, reuse_weights=True, weights=fn)
    assert regridder_reuse.weights.shape == regridder.weights.shape

    # or can also overwrite it
    xe.Regridder(ds_in, ds_out, method)

    # check legacy args still work
    regridder = xe.Regridder(ds_in, ds_out, method, filename='wgts.nc')
    regridder_reuse = xe.Regridder(ds_in, ds_out, method, reuse_weights=True, filename='wgts.nc')
    assert regridder_reuse.weights.shape == regridder.weights.shape

    # check fails on non-existent file
    with pytest.raises(OSError):
        xe.Regridder(ds_in, ds_out, method, reuse_weights=True, filename='fakewgts.nc')

    # check fails if no weights are provided
    with pytest.raises(ValueError):
        xe.Regridder(ds_in, ds_out, method, reuse_weights=True)


def test_regridder_w():
    """Check that the `w` property for dimensioned weights works."""
    regridder = xe.Regridder(ds_in, ds_out, method='bilinear')
    w = regridder.w
    assert w.shape == ds_out.lon.shape + ds_in.lon.shape

    p = segmentize(Polygon([(-10, -10), (10, -10), (10, 10), (-10, 10)]), 1)

    averager = xe.SpatialAverager(ds_in, [p])
    assert averager.w.shape == (1,) + ds_in.lon.shape

    ds_in_cf = xe.util.grid_global(15, 15, cf=True)
    ds_out_cf = xe.util.grid_global(30, 30, cf=True)

    regridder_cf = xe.Regridder(ds_in_cf, ds_out_cf, method='bilinear')
    w_cf = regridder_cf.w
    assert w_cf.shape == (
        ds_out_cf.lat.shape + ds_out_cf.lon.shape + ds_in_cf.lat.shape + ds_in_cf.lon.shape
    )

    averager = xe.SpatialAverager(ds_in_cf, [p, p])
    assert averager.w.shape == (2,) + ds_in_cf.lat.shape + ds_in_cf.lon.shape


def test_to_netcdf(tmp_path):
    from xesmf.backend import Grid, esmf_regrid_build

    # Let the frontend write the weights to disk
    xfn = tmp_path / 'ESMF_weights.nc'
    method = 'bilinear'
    regridder = xe.Regridder(ds_in, ds_out, method, unmapped_to_nan=False)
    regridder.to_netcdf(filename=xfn)

    grid_in = Grid.from_xarray(ds_in['lon'].values.T, ds_in['lat'].values.T)
    grid_out = Grid.from_xarray(ds_out['lon'].values.T, ds_out['lat'].values.T)

    # Let the ESMPy backend write the weights to disk
    efn = tmp_path / 'weights.nc'
    esmf_regrid_build(grid_in, grid_out, method=method, filename=str(efn))

    x = xr.open_dataset(xfn)
    e = xr.open_dataset(efn)
    xr.testing.assert_identical(x, e)


def test_to_netcdf_nans(tmp_path):
    from xesmf.backend import Grid, esmf_regrid_build

    # Let the frontend write the weights to disk
    xfn = tmp_path / 'ESMF_weights_nans.nc'
    method = 'bilinear'
    regridder = xe.Regridder(ds_in, ds_out, method, unmapped_to_nan=True)
    regridder.to_netcdf(filename=xfn)

    grid_in = Grid.from_xarray(ds_in['lon'].values.T, ds_in['lat'].values.T)
    grid_out = Grid.from_xarray(ds_out['lon'].values.T, ds_out['lat'].values.T)

    # Let the ESMPy backend write the weights to disk
    efn = tmp_path / 'weights_nans.nc'
    esmf_regrid_build(grid_in, grid_out, method=method, filename=str(efn))

    x = xr.open_dataset(xfn)
    e = xr.open_dataset(efn)

    # Reformat to sparse COO matrix
    smat = xe.smm.read_weights(e, np.prod(ds_in['lon'].shape), np.prod(ds_out['lon'].shape))
    # Add NaNs to weights
    smat = xe.smm.add_nans_to_weights(smat)
    # Updating the dataset
    e_nans = xr.Dataset(
        {
            'S': (['n_s'], smat.data.data),
            'row': (['n_s'], smat.data.coords[0, :] + 1),
            'col': (['n_s'], smat.data.coords[1, :] + 1),
        }
    )
    # Comparison
    xr.testing.assert_identical(x, e_nans)


def test_conservative_without_bounds():
    with pytest.raises(KeyError):
        xe.Regridder(ds_in.drop_vars('lon_b'), ds_out, 'conservative')


def test_regridder_from_dict():
    lon_in = ds_in['lon'].values
    lat_in = ds_in['lat'].values
    lon_out = ds_out['lon'].values
    lat_out = ds_out['lat'].values
    reg = xe.Regridder({'lon': lon_in, 'lat': lat_in}, {'lon': lon_out, 'lat': lat_out}, 'bilinear')
    with pytest.warns(UserWarning, match=r"Using dimensions \('y', 'x'\) from data variable"):
        reg(ds_in['data'])


def test_regrid_periodic_wrong():
    # not using periodic option
    regridder = xe.Regridder(ds_in, ds_out, 'bilinear', unmapped_to_nan=False)

    dr_out = regridder(ds_in['data'])  # xarray DataArray

    # compare with analytical solution
    rel_err = (ds_out['data_ref'] - dr_out) / ds_out['data_ref']
    assert np.max(np.abs(rel_err)) == 1.0  # some data will be missing


def test_regrid_periodic_correct():
    regridder = xe.Regridder(ds_in, ds_out, 'bilinear', periodic=True)

    dr_out = regridder(ds_in['data'])

    # compare with analytical solution
    rel_err = (ds_out['data_ref'] - dr_out) / ds_out['data_ref']
    assert np.max(np.abs(rel_err)) < 0.065


def ds_2d_to_1d(ds):
    ds_temp = ds.reset_coords()
    ds_1d = xr.merge([ds_temp['lon'][0, :], ds_temp['lat'][:, 0]])
    ds_1d.coords['lon'] = ds_1d['lon']
    ds_1d.coords['lat'] = ds_1d['lat']
    return ds_1d


@pytest.mark.parametrize('dtype', ['float32', 'float64'])
@pytest.mark.parametrize(
    'data_in',
    [
        pytest.param(np.array(ds_in['data']), id='np.ndarray'),
        pytest.param(xr.DataArray(ds_in['data']), id='xr.DataArray input'),
        pytest.param(xr.Dataset(ds_in[['data']]), id='xr.Dataset input'),
        pytest.param(ds_in['data'].chunk(), id='da.Array input'),
    ],
)
def test_regridded_respects_input_dtype(dtype, data_in):
    """Tests regridded output has same dtype as input"""
    data_in = data_in.astype(dtype)
    regridder = xe.Regridder(ds_in, ds_out, 'bilinear')  # Make this a fixture?
    out = regridder(data_in)

    if isinstance(data_in, xr.Dataset):
        # When data_in is xr.Dataset, a mapping...
        assert out['data'].dtype == data_in['data'].dtype
    else:
        assert out.dtype == data_in.dtype


def test_regrid_with_1d_grid():
    ds_in_1d = ds_2d_to_1d(ds_in)
    ds_out_1d = ds_2d_to_1d(ds_out)

    regridder = xe.Regridder(ds_in_1d, ds_out_1d, 'bilinear', periodic=True)

    dr_out = regridder(ds_in['data'])

    # compare with analytical solution
    rel_err = (ds_out['data_ref'] - dr_out) / ds_out['data_ref']
    assert np.max(np.abs(rel_err)) < 0.065

    # metadata should be 1D
    assert_equal(dr_out['lon'].values, ds_out_1d['lon'].values)
    assert_equal(dr_out['lat'].values, ds_out_1d['lat'].values)


def test_regrid_with_1d_grid_infer_bounds():
    ds_in_1d = ds_2d_to_1d(ds_in).swap_dims(x='lon', y='lat')
    ds_out_1d = ds_2d_to_1d(ds_out).swap_dims(x='lon', y='lat')

    regridder = xe.Regridder(ds_in_1d, ds_out_1d, 'conservative', periodic=True)
    with pytest.warns(UserWarning, match=r"Using dimensions \('y', 'x'\) from data variable"):
        dr_out = regridder(ds_in['data'])

    # compare with provided-bounds solution
    dr_exp = xe.Regridder(ds_in, ds_out, 'conservative', periodic=True)(ds_in['data'])

    assert_allclose(dr_out, dr_exp)


def test_regrid_cfbounds():
    # Test regridding when bounds are given in cf format with a custom "bounds" name.
    ds = ds_in.copy().drop_vars(['lat_b', 'lon_b'])
    ds['lon_bounds'] = cfxr.vertices_to_bounds(ds_in.lon_b, ('bnds', 'y', 'x'))
    ds['lat_bounds'] = cfxr.vertices_to_bounds(ds_in.lat_b, ('bnds', 'y', 'x'))
    ds.lat.attrs['bounds'] = 'lat_bounds'
    ds.lon.attrs['bounds'] = 'lon_bounds'

    regridder = xe.Regridder(ds, ds_out, 'conservative', periodic=True)
    dr_out = regridder(ds['data'])
    # compare with provided-bounds solution
    dr_exp = xe.Regridder(ds_in, ds_out, 'conservative', periodic=True)(ds_in['data'])
    assert_allclose(dr_out, dr_exp)


# TODO: consolidate (regrid method, input data types) combination
# using pytest fixtures and parameterization


@pytest.mark.parametrize('use_cfxr', [True, False])
def test_regrid_dataarray(use_cfxr):
    # xarray.DataArray containing in-memory numpy array
    if use_cfxr:
        ds_in2 = ds_in.rename(lat='latitude', lon='longitude')
        ds_out2 = ds_out.rename(lat='latitude', lon='longitude')
    else:
        ds_in2 = ds_in
        ds_out2 = ds_out

    regridder = xe.Regridder(ds_in2, ds_out2, 'conservative')

    outdata = regridder(ds_in2['data'].values)  # pure numpy array
    dr_out = regridder(ds_in2['data'])  # xarray DataArray

    # DataArray and numpy array should lead to the same result
    assert_equal(outdata, dr_out.values)

    # compare with analytical solution
    rel_err = (ds_out2['data_ref'] - dr_out) / ds_out2['data_ref']
    assert np.max(np.abs(rel_err)) < 0.05

    # check metadata
    lat_name = 'latitude' if use_cfxr else 'lat'
    lon_name = 'longitude' if use_cfxr else 'lon'
    xr.testing.assert_identical(dr_out[lat_name], ds_out2[lat_name])
    xr.testing.assert_identical(dr_out[lon_name], ds_out2[lon_name])

    # test broadcasting
    dr_out_4D = regridder(ds_in2['data4D'])

    # data over broadcasting dimensions should agree
    assert_almost_equal(
        ds_in2['data4D'].values.mean(axis=(2, 3)),
        dr_out_4D.values.mean(axis=(2, 3)),
        decimal=10,
    )

    # check metadata
    xr.testing.assert_identical(dr_out_4D['time'], ds_in2['time'])
    xr.testing.assert_identical(dr_out_4D['lev'], ds_in2['lev'])

    # test transposed
    dr_out_4D_t = regridder(ds_in2['data4D'].transpose(..., 'time', 'lev'))
    xr.testing.assert_identical(dr_out_4D, dr_out_4D_t)

    # test renamed dim
    if not use_cfxr:
        with pytest.warns(UserWarning, match=r"Using dimensions \('why', 'x'\)"):
            dr_out_rn = regridder(ds_in2.rename(y='why')['data'])
        xr.testing.assert_identical(dr_out, dr_out_rn)


@pytest.mark.parametrize('use_dask', [True, False])
def test_regrid_dataarray_endianess(use_dask):
    # xarray.DataArray containing in-memory numpy array
    regridder = xe.Regridder(ds_in, ds_out, 'conservative')

    exp = regridder(ds_in['data'])  # Normal (little-endian)

    if use_dask:
        indata = ds_in.data.astype('>f8').chunk()
    else:
        indata = ds_in.data.astype('>f8')

    with pytest.warns(UserWarning, match='Input array has a dtype not supported'):
        out = regridder(indata)  # big endian

    # Results should be the same
    assert_equal(exp.values, out.values)
    assert out.dtype == '>f8'


def test_regrid_from_dataarray():
    # Check that creating Regridder from DataArrays doesn't fail
    xe.Regridder(ds_in['data'], ds_out['data_ref'], 'bilinear')


def test_regrid_dataarray_to_locstream():
    # xarray.DataArray containing in-memory numpy array

    regridder = xe.Regridder(ds_in, ds_locs, 'bilinear', locstream_out=True)

    outdata = regridder(ds_in['data'].values)  # pure numpy array
    dr_out = regridder(ds_in['data'])  # xarray DataArray
    dr_out_t = regridder(ds_in['data'].transpose())  # Transpose to test name matching

    # DataArray and numpy array should lead to the same result
    assert_equal(outdata.squeeze(), dr_out.values)
    assert_equal(outdata.squeeze(), dr_out_t.values)

    with pytest.raises(ValueError):
        regridder = xe.Regridder(ds_in, ds_locs, 'conservative', locstream_out=True)


def test_regrid_dataarray_from_locstream():
    # xarray.DataArray containing in-memory numpy array

    regridder = xe.Regridder(ds_locs, ds_in, 'nearest_s2d', locstream_in=True)

    outdata = regridder(ds_locs['lat'].values)  # pure numpy array
    dr_out = regridder(ds_locs['lat'])  # xarray DataArray
    # New dim and transpose to test name-matching
    dr_out_2D = regridder(ds_locs['lat'].expand_dims(other=[1]).transpose('locations', 'other'))

    # DataArray and numpy array should lead to the same result
    assert_equal(outdata, dr_out.values)
    assert_equal(outdata, dr_out_2D.sel(other=1).values)

    with pytest.raises(ValueError):
        regridder = xe.Regridder(ds_locs, ds_in, 'bilinear', locstream_in=True)
    with pytest.raises(ValueError):
        regridder = xe.Regridder(ds_locs, ds_in, 'patch', locstream_in=True)
    with pytest.raises(ValueError):
        regridder = xe.Regridder(ds_locs, ds_in, 'conservative', locstream_in=True)


@pytest.mark.parametrize('scheduler', dask_schedulers)
def test_regrid_dask(request, scheduler):
    # chunked dask array (no xarray metadata)
    scheduler = request.getfixturevalue(scheduler)
    regridder = xe.Regridder(ds_in, ds_out, 'conservative')

    indata = ds_in_chunked['data'].data
    outdata = regridder(indata)

    assert dask.is_dask_collection(outdata)

    # lazy dask arrays have incorrect shape attribute due to last chunk
    assert outdata.shape == indata.shape[:-2] + horiz_shape_out

    # Check that the number of tasks is as predicted
    # ds_in has 1 chunk
    # thus output also has 1 chunk (output is not chunked if input isn't)
    # regridding adds 3 tasks, wrapping the weights adds 2
    n_task_out = len(outdata.__dask_graph__().keys())
    n_task_in = len(indata.__dask_graph__().keys())
    assert n_task_out == n_task_in + 5

    # Use very small chunks
    indata_chunked = indata.rechunk((5, 6))  # Now has 9 chunks (5, 6)
    with pytest.warns(
        PerformanceWarning,
        match=r'Regridding is increasing the number of chunks by a factor of 16.0',
    ):
        outdata = regridder(indata_chunked)
    # This is the case where we preserve chunk size
    assert outdata.chunksize == indata_chunked.chunksize
    n_task_out = len(outdata.__dask_graph__().keys())
    n_task_in = len(indata_chunked.__dask_graph__().keys())
    # input has 9 chunks
    # output has 16
    # Regridding adds 2 * 9 * 16 + 16 + 64 (I'm not sure I fully understand how dasks sums at the end)
    # Wrapping the weights adds 9 * 16 + 1
    assert n_task_out == n_task_in + 513

    # Prescribe chunks
    outdata = regridder(indata, output_chunks=(-1, 12))
    n_task_out = len(outdata.__dask_graph__().keys())
    n_task_in = len(indata.__dask_graph__().keys())
    # input has 1 chunks
    # output has 2
    # Regridding adds 2 * 1 * 2 + 2
    # Wrapping the weights adds 1 * 2 + 1
    assert n_task_out == n_task_in + 9

    outdata_ref = ds_out['data_ref'].values
    rel_err = (outdata.compute() - outdata_ref) / outdata_ref
    assert np.max(np.abs(rel_err)) < 0.05


@pytest.mark.parametrize('scheduler', dask_schedulers)
def test_regrid_dask_to_locstream(request, scheduler):
    # chunked dask array (no xarray metadata)

    scheduler = request.getfixturevalue(scheduler)
    regridder = xe.Regridder(ds_in, ds_locs, 'bilinear', locstream_out=True)

    indata = ds_in_chunked['data4D'].data
    outdata = regridder(indata)
    assert dask.is_dask_collection(outdata)


@pytest.mark.parametrize('scheduler', dask_schedulers)
def test_regrid_dask_from_locstream(request, scheduler):
    # chunked dask array (no xarray metadata)

    scheduler = request.getfixturevalue(scheduler)
    regridder = xe.Regridder(ds_locs, ds_in, 'nearest_s2d', locstream_in=True)

    outdata = regridder(ds_locs.chunk()['lat'].data)
    assert dask.is_dask_collection(outdata)


@pytest.mark.parametrize('scheduler', dask_schedulers)
def test_regrid_dataarray_dask(request, scheduler):
    # xarray.DataArray containing chunked dask array
    scheduler = request.getfixturevalue(scheduler)
    regridder = xe.Regridder(ds_in, ds_out, 'conservative')

    dr_in = ds_in_chunked['data4D']
    dr_out = regridder(dr_in)
    assert dask.is_dask_collection(dr_out)

    assert dr_out.data.shape == dr_in.data.shape[:-2] + horiz_shape_out

    # data over broadcasting dimensions should agree
    assert_almost_equal(dr_in.values.mean(axis=(2, 3)), dr_out.values.mean(axis=(2, 3)), decimal=10)

    # check metadata
    xr.testing.assert_identical(dr_out['time'], dr_in['time'])
    xr.testing.assert_identical(dr_out['lev'], dr_in['lev'])
    assert_equal(dr_out['lat'].values, ds_out['lat'].values)
    assert_equal(dr_out['lon'].values, ds_out['lon'].values)


@pytest.mark.parametrize('scheduler', dask_schedulers)
def test_regrid_dataarray_dask_to_locstream(request, scheduler):
    # xarray.DataArray containing chunked dask array
    scheduler = request.getfixturevalue(scheduler)
    regridder = xe.Regridder(ds_in, ds_locs, 'bilinear', locstream_out=True)

    dr_in = ds_in_chunked['data4D']
    dr_out = regridder(dr_in)
    assert dask.is_dask_collection(dr_out)


@pytest.mark.parametrize('scheduler', dask_schedulers)
def test_regrid_dataarray_dask_from_locstream(request, scheduler):
    # xarray.DataArray containing chunked dask array

    scheduler = request.getfixturevalue(scheduler)
    regridder = xe.Regridder(ds_locs, ds_in, 'nearest_s2d', locstream_in=True)

    outdata = regridder(ds_locs.chunk()['lat'])
    assert dask.is_dask_collection(outdata)


def test_dask_output_chunks():
    regridder = xe.Regridder(ds_in, ds_out, 'conservative')

    test_output_chunks_tuple = (10, 12)
    test_output_chunks_dict = {'y': 10, 'x': 12}
    indata = ds_spatial_chunked['data4D'].data  # Data chunked along spatial dims
    # Use ridiculous small chunk size value to be sure it _isn't_ impacting computation.
    with dask.config.set({'array.chunk-size': '1MiB'}), pytest.warns(PerformanceWarning):
        outdata = regridder(indata)
        outdata_spec_tuple = regridder(indata, output_chunks=test_output_chunks_tuple)
        outdata_spec_dict = regridder(indata, output_chunks=test_output_chunks_dict)

    assert dask.is_dask_collection(outdata)
    assert dask.is_dask_collection(outdata_spec_tuple)
    assert dask.is_dask_collection(outdata_spec_dict)

    # Verify that the default chunking is correct
    assert outdata.shape == indata.shape[:-2] + horiz_shape_out
    assert outdata.chunksize == indata.chunksize

    # Verify that we get specified outputchunks when the argument is provided
    assert outdata_spec_tuple.shape == indata.shape[:-2] + horiz_shape_out
    assert outdata_spec_tuple.chunksize == indata.chunksize[:-2] + test_output_chunks_tuple

    assert outdata_spec_dict.shape == indata.shape[:-2] + horiz_shape_out
    assert (
        outdata_spec_dict.chunksize == indata.chunksize[:-2] + test_output_chunks_tuple
    )  # dict should've been converted to tuple


def test_para_weight_gen():
    # Generating weights in serial and parallel
    regridder = xe.Regridder(ds_in, ds_out, 'conservative')
    para_regridder = xe.Regridder(ds_in, ds_out_chunked, 'conservative', parallel=True)

    # weights should be identical between serial and parallel
    assert all(regridder.w.data.data == para_regridder.w.data.data)

    # Should work with a rectilinear version too (where dims == coords)
    ds_in_cf = xe.util.cf_grid_2d(-90, 90, 20, -45, 45, 12)
    ds_out_cf = xe.util.cf_grid_2d(-90, 90, 15, -45, 45, 9)
    ds_in_cf['data'] = xe.data.wave_smooth(ds_in_cf['lon'], ds_in_cf['lat'])
    ds_out_cf['data_ref'] = xe.data.wave_smooth(ds_out_cf['lon'], ds_out_cf['lat']).chunk(
        {'lat': 5, 'lon': 5}
    )

    # Generating weights in serial and parallel
    regridder = xe.Regridder(ds_in_cf, ds_out_cf, 'conservative')
    para_regridder = xe.Regridder(ds_in_cf, ds_out_cf, 'conservative', parallel=True)

    # weights should be identical between serial and parallel
    assert all(regridder.w.data.data == para_regridder.w.data.data)

    # Ensure para weight gen works with locstream_in as well
    reggrider_locs = xe.Regridder(ds_locs, ds_out_chunked, 'nearest_s2d', locstream_in=True)
    para_regridder_locs = xe.Regridder(
        ds_locs, ds_out_chunked, 'nearest_s2d', parallel=True, locstream_in=True
    )
    assert all(reggrider_locs.w.data.data == para_regridder_locs.w.data.data)

    # Same as above with locstream_out
    regridder_locs = xe.Regridder(ds_in, ds_locs, 'nearest_s2d', locstream_out=True)
    para_regridder_locs = xe.Regridder(
        ds_in, ds_locs_chunked, 'nearest_s2d', parallel=True, locstream_out=True
    )
    assert all(regridder_locs.w.data.data == para_regridder_locs.w.data.data)


def test_para_weight_gen_errors():
    with pytest.raises(ValueError, match='requires the output grid to have chunks'):
        xe.Regridder(ds_in, ds_out, 'conservative', parallel=True)


def test_regrid_dataset():
    # xarray.Dataset containing in-memory numpy array
    regridder = xe.Regridder(ds_in, ds_out, 'conservative')

    # `ds_out` already refers to output grid object
    # TODO: use more consistent variable namings across tests
    ds_result = regridder(ds_in)

    # output should contain all data variables
    assert set(ds_result.data_vars.keys()) == set(ds_in.data_vars.keys())

    # compare with analytical solution
    rel_err = (ds_out['data_ref'] - ds_result['data']) / ds_out['data_ref']
    assert np.max(np.abs(rel_err)) < 0.05

    # data over broadcasting dimensions should agree
    assert_almost_equal(
        ds_in['data4D'].values.mean(axis=(2, 3)),
        ds_result['data4D'].values.mean(axis=(2, 3)),
        decimal=10,
    )

    assert ds_result['data4D'].dtype == np.dtype('f8')
    assert ds_result['data4D_f4'].dtype == np.dtype('f4')

    # check metadata
    xr.testing.assert_identical(ds_result['time'], ds_in['time'])
    xr.testing.assert_identical(ds_result['lev'], ds_in['lev'])
    assert_equal(ds_result['lat'].values, ds_out['lat'].values)
    assert_equal(ds_result['lon'].values, ds_out['lon'].values)

    # Allow (but skip) other non spatial variables
    ds_result2 = regridder(ds_in.assign(nonspatial=ds_in.x * ds_in.time))
    xr.testing.assert_identical(ds_result2, ds_result)


def test_regrid_dataset_extracoords():
    ds_out2 = ds_out.copy().assign_coords(
        x=np.arange(24),
        y=np.arange(20),  # coords to be transfered
        latitude_longitude=xr.DataArray(),  # grid_mapping
        bogus=ds_out.lev * ds_out.lon,  # coords not to be transfered
        scalar1=1,  #
        scalar2=1,  #
    )
    ds_out2['data_ref'].attrs['grid_mapping'] = 'latitude_longitude'
    ds_out2['data4D_ref'].attrs['grid_mapping'] = 'latitude_longitude'

    ds_in2 = ds_in.assign_coords(scalar2=5)
    regridder = xe.Regridder(ds_in, ds_out2, 'conservative')
    ds_result = regridder(ds_in2)

    assert 'x' in ds_result.coords
    assert 'y' in ds_result.coords
    assert 'bogus' not in ds_result.coords
    assert 'scalar1' not in ds_result.coords
    assert ds_result.scalar2 == 5
    assert 'latitude_longitude' in ds_result.coords


@pytest.mark.parametrize('scheduler', dask_schedulers)
def test_regrid_dataset_dask(request, scheduler):
    scheduler = request.getfixturevalue(scheduler)
    # xarray.Dataset containing dask array
    regridder = xe.Regridder(ds_in, ds_out, 'conservative')

    # `ds_out` already refers to output grid object
    ds_result = regridder(ds_in.chunk())

    # output should contain all data variables
    assert set(ds_result.data_vars.keys()) == set(ds_in.data_vars.keys())
    assert dask.is_dask_collection(ds_result)
    assert ds_result.data.dtype == ds_in.data.dtype

    ds_in_f4 = ds_in.copy()
    ds_in_f4['data'] = ds_in_f4.data.astype('float32')
    ds_in_f4['data4D'] = ds_in_f4.data4D.astype('float32')
    ds_result = regridder(ds_in_f4.chunk())
    assert ds_result.data.dtype == 'float32'


def test_regrid_dataset_to_locstream():
    # xarray.Dataset containing in-memory numpy array

    regridder = xe.Regridder(ds_in, ds_locs, 'bilinear', locstream_out=True)
    regridder(ds_in)


@pytest.mark.parametrize('transpose', [True, False])
def test_build_regridder_with_masks(transpose):
    dsi = ds_in.copy()
    mask = xr.DataArray(np.random.randint(2, size=ds_in['data'].shape), dims=('y', 'x'))
    if transpose:
        mask = mask.T
    dsi['mask'] = mask
    # 'patch' is too slow to test
    for method in [
        'bilinear',
        'conservative',
        'conservative_normed',
        'nearest_s2d',
        'nearest_d2s',
    ]:
        regridder = xe.Regridder(dsi, ds_out, method)

        # check screen output
        assert repr(regridder) == str(regridder)
        assert 'xESMF Regridder' in str(regridder)
        assert method in str(regridder)


def test_regrid_dataset_from_locstream():
    # xarray.Dataset containing in-memory numpy array

    regridder = xe.Regridder(ds_locs, ds_in, 'nearest_s2d', locstream_in=True)
    regridder(ds_locs)


def test_ds_to_ESMFlocstream():
    try:
        import esmpy as ESMF
    except ImportError:
        import ESMF

    from xesmf.frontend import ds_to_ESMFlocstream

    locstream, shape, names = ds_to_ESMFlocstream(ds_locs)
    assert isinstance(locstream, ESMF.LocStream)
    assert shape == (
        1,
        4,
    )
    assert names == ('locations',)
    with pytest.raises(ValueError):
        locstream, shape, names = ds_to_ESMFlocstream(ds_in)
    ds_bogus = ds_in.copy()
    ds_bogus['lon'] = ds_locs['lon']
    with pytest.raises(ValueError):
        locstream, shape, names = ds_to_ESMFlocstream(ds_bogus)


@pytest.mark.parametrize('use_dask', [True, False])
@pytest.mark.parametrize('poly,exp', list(zip(polys, exps_polys)))
def test_spatial_averager(poly, exp, use_dask):
    if isinstance(poly, (Polygon, MultiPolygon)):
        poly = [poly]
    if use_dask:
        ds_in = ds_savg.chunk(lat=10)
    else:
        ds_in = ds_savg
    savg = xe.SpatialAverager(ds_in, poly, geom_dim_name='my_geom')
    out = savg(ds_in.abc)
    assert_allclose(out, exp, rtol=1e-3)

    assert 'my_geom' in out.dims


def test_spatial_averager_warns():
    with pytest.warns(UserWarning, match=r'contains large \(> 1°\) segments.'):
        xe.SpatialAverager(ds_savg, [polys_raw[0]], geom_dim_name='my_geom')


def test_spatial_averager_with_zonal_region():
    # We expect the spatial average for all regions to be one
    zonal_south = Polygon([(0, -90), (10, 0), (0, 0)])
    zonal_north = Polygon([(0, 90), (10, 0), (0, 0)])
    zonal_short = Polygon([(0, -10), (10, -10), (10, 10), (0, 10)])
    zonal_full = Polygon([(0, -90), (10, 0), (0, 90), (0, 0)])  # This yields 0... why?

    polys = [zonal_south, zonal_north, zonal_short, zonal_full]
    polys = segmentize(polys, 0.9)

    # Create field of ones on a global grid
    ds = xe.util.grid_global(20, 12, cf=True)
    ds['a'] = xr.DataArray(
        np.ones((ds.lon.size, ds.lat.size)),
        coords={'lat': ds.lat, 'lon': ds.lon},
        dims=('lon', 'lat'),
    )
    out = xe.SpatialAverager(ds, polys)(ds.a)
    assert_allclose(out, 1, rtol=1e-3)


@pytest.mark.filterwarnings('ignore:`polys` contains large')
def test_compare_weights_from_poly_and_grid():
    """Confirm that the weights are identical when they are computed from a grid->grid and grid->poly."""
    pytest.importorskip(
        'cf_xarray', minversion='0.10.9', reason='cf-xarray 0.10.8 broken for singleton coordinates'
    )
    # Global grid
    ds = xe.util.grid_global(20, 12, cf=True)

    # A single destination tile
    tile = xe.util.cf_grid_2d(-40, -80, -40, 0, -80, -80)
    ds['a'] = xr.DataArray(
        np.ones((ds.lon.size, ds.lat.size)),
        coords={'lat': ds.lat, 'lon': ds.lon},
        dims=('lon', 'lat'),
    )

    # Create polygon from tile corners
    x1, x2 = tile.lon_bounds.isel(lon=0)
    y1, y2 = tile.lat_bounds.isel(lat=0)
    poly = Polygon([(x1, y1), (x2, y1), (x2, y2), (x1, y2)])

    # Regrid using two identical destination grids (in theory)
    rgrid = xe.Regridder(ds, tile, method='conservative')
    rpoly = xe.SpatialAverager(ds, [poly])

    # Visualize the weights
    wg = np.reshape(rgrid.weights.data.todense(), ds.a.T.shape)
    wp = np.reshape(rpoly.weights.data.todense(), ds.a.T.shape)

    # Normally, weights should be identical, but this fails
    np.testing.assert_array_almost_equal(wg, wp)

    # Check that source area affects weights
    ds['wg'] = (('lat', 'lon'), wg)
    ds['wp'] = (('lat', 'lon'), wp)
    # i.e. weights are larger for a cell closer to the equator, considering two completely covered cells
    assert ds.wg.sel(lon=-70, lat=-60) > ds.wg.sel(lon=-70, lat=-72)


def test_polys_to_ESMFmesh():
    try:
        import esmpy as ESMF
    except ImportError:
        import ESMF

    from xesmf.frontend import polys_to_ESMFmesh

    # No overlap but multi + holes
    with warnings.catch_warnings(record=True) as rec:
        mesh, shape = polys_to_ESMFmesh([polys[1], polys[2], polys[4]])

    assert isinstance(mesh, ESMF.Mesh)
    assert shape == (1, 4)
    assert len(rec) == 1
    assert 'Some passed polygons have holes' in rec[0].message.args[0]


@pytest.mark.parametrize(
    'method, skipna, na_thres, nvalid',
    [
        ('bilinear', False, 1.0, 380),
        ('bilinear', True, 1.0, 395),
        ('bilinear', True, 0.0, 380),
        ('bilinear', True, 0.5, 388),
        ('bilinear', True, 1.0, 395),
        ('conservative', False, 1.0, 385),
        ('conservative', True, 1.0, 394),
        ('conservative', True, 0.0, 385),
        ('conservative', True, 0.5, 388),
        ('conservative', True, 1.0, 394),
    ],
)
def test_skipna(method, skipna, na_thres, nvalid):
    dai = ds_in['data4D'].copy()
    dai[0, 0, 4:6, 4:6] = np.nan
    rg = xe.Regridder(ds_in, ds_out, method)
    dao = rg(dai, skipna=skipna, na_thres=na_thres)
    assert int(dao[0, 0, 1:-1, 1:-1].notnull().sum()) == nvalid


def test_non_cf_latlon():
    ds_in_noncf = ds_in.copy()
    ds_in_noncf.lon.attrs = {}
    ds_in_noncf.lat.attrs = {}
    # Test non-CF lat/lon extraction for both DataArray and Dataset
    xe.Regridder(ds_in_noncf['data'], ds_out, 'bilinear')
    xe.Regridder(ds_in_noncf, ds_out, 'bilinear')


@pytest.mark.parametrize(
    'var_renamer,dim_out',
    [
        ({}, 'locations'),
        ({'lon': {'locations': 'foo'}, 'lat': {'locations': 'foo'}}, 'foo'),
        ({'lon': {'locations': 'foo'}, 'lat': {'locations': 'bar'}}, None),
    ],
)
def test_locstream_dim_name(var_renamer, dim_out):
    ds_locs_renamed = ds_locs.copy()
    for var, renamer in var_renamer.items():
        ds_locs_renamed[var] = ds_locs_renamed[var].rename(renamer)

    if dim_out is None:
        with pytest.raises(ValueError, match='not specified along the same dimension'):
            regridder = xe.Regridder(ds_in, ds_locs_renamed, 'bilinear', locstream_out=True)
    else:
        regridder = xe.Regridder(ds_in, ds_locs_renamed, 'bilinear', locstream_out=True)
        expected = {'lev', 'time', 'x_b', 'y_b', dim_out}
        actual = set(regridder(ds_in).dims)
        assert expected == actual


def test_spatial_averager_mask():
    poly = Polygon([[0.5, 0.5], [1.5, 0.5], [1.5, 1.5], [0.5, 1.5]])
    ds = ds_savg.copy(deep=True)
    ds.abc[1, 1] = np.nan
    savg = xe.SpatialAverager(ds, [poly], geom_dim_name='my_geom')

    # Without mask, we expect NaNs to propagate
    out = savg(ds.abc)
    assert out.isnull()

    # With masking, the NaN should be ignored.
    mask = ds.abc.notnull()
    dsm = ds.assign(
        mask=mask.T
    )  # TODO: open an issue about the fact that this fails without a transpose.
    savg = xe.SpatialAverager(dsm, [poly], geom_dim_name='my_geom')
    out = savg(dsm.abc)
    assert_allclose(out, 2, rtol=1e-3)


def test_post_mask_source():
    # Define source and target grids
    ds_in = xr.Dataset(
        {'lat': (['lat'], np.linspace(1, 10, 10)), 'lon': (['lon'], np.linspace(1, 10, 10))}
    )
    ds_out = xr.Dataset(
        {'lat': (['lat'], np.linspace(0, 15, 20)), 'lon': (['lon'], np.linspace(0, 15, 20))}
    )

    # Build regridder with post_source_mask='domain_edge'
    regridder = xe.Regridder(ds_in, ds_out, 'nearest_s2d', post_mask_source='domain_edge')

    # Regridding
    da_in = xr.DataArray(np.ones((10, 10)), dims=['lat', 'lon'])
    da_in[0, :] = 10  # top edge
    da_in[-1, :] = 10  # bottom edge
    da_out = regridder(da_in)

    # Check that edge values (10) did not make it into the result
    assert not np.any(da_out.values > 1), 'Edge contributions were not removed'

    # Check that cells outside the original domain are 0
    lat_mask = (ds_out.lat >= 10) | (ds_out.lat <= 1)
    lon_mask = (ds_out.lon >= 10) | (ds_out.lon <= 1)
    assert np.all(da_out.sel(lat=lat_mask).values == 0)
    assert np.all(da_out.sel(lon=lon_mask).values == 0)


def test_post_mask_source_exceptions():
    # LocStream source should give an exception
    ds_src = xr.Dataset(
        {
            'lat': (['location'], np.linspace(-90, 90, 5)),
            'lon': (['location'], np.linspace(-180, 180, 5)),
        }
    )
    ds_dst = xe.util.grid_global(5, 5)
    with pytest.raises(ValueError, match="post_mask_source='domain_edge' is only supported.*"):
        xe.Regridder(
            ds_src, ds_dst, 'nearest_s2d', post_mask_source='domain_edge', locstream_in=True
        )

    # Not using an array-like should give an exception
    ds_src = xe.util.grid_global(10, 10)

    class NotArrayLike:
        def __array__(self, dtype=None):
            raise ValueError

    with pytest.raises(TypeError, match='must be array-like of integers'):
        xe.Regridder(ds_src, ds_dst, 'bilinear', post_mask_source=NotArrayLike())

    # Not using integer indices should give an exception
    bad_index_mask = np.array([0.1, 2.5, 3.7])
    with pytest.raises(TypeError, match='must be of integer type'):
        xe.Regridder(ds_src, ds_dst, 'bilinear', post_mask_source=bad_index_mask)

    # Using out of bounds indices should give an exception
    bad_index_mask = np.array([100, 200, 300, 400, 500, 600, 700])
    with pytest.raises(ValueError, match='indices are out of valid range \\[0, 648\\)'):
        xe.Regridder(ds_src, ds_dst, 'bilinear', post_mask_source=bad_index_mask)


def test_post_mask_source_parallel_mode():
    # Create source and destination grids
    ds_src = xe.util.grid_global(10, 10)
    ds_dst = xe.util.grid_global(15, 15)

    # Chunk destination data to enable parallel mode
    mask = np.ones((12, 24))
    ds_dst['mask'] = xr.DataArray(data=mask, dims=('y', 'x'))
    ds_dst = ds_dst.chunk({'y': 12, 'x': 12})

    # Define source data with known values
    data = xr.DataArray(
        np.ones((18, 36)),
        dims=['y', 'x'],
        coords={'lat': ds_src['lat'], 'lon': ds_src['lon']},
    )

    # Choose some source indices to mask (e.g., corners)
    mask_indices = np.array([0, 1, 2, 3, 4])

    # Create regridder with post_mask_source and parallel=True
    regridder = xe.Regridder(
        ds_src,
        ds_dst,
        'nearest_s2d',
        post_mask_source=mask_indices,
        parallel=True,
        unmapped_to_nan=True,
    )

    # Regrid and compute the result
    result = regridder(data).compute()

    # Assert that the result contains NaNs
    assert result.shape == (12, 24)
    assert isinstance(result, xr.DataArray)
    assert result.isnull().any()


def test_locstream_input_grid_output_with_target_mask_applied():
    # Create locstream input (6 coordinate points)
    locstream_in = xr.Dataset(
        {'var': xr.DataArray(np.ones((6)), dims=['location'])},
        coords={
            'lat': ('location', np.linspace(0, 5, 6)),
            'lon': ('location', np.linspace(0, 10, 6)),
        },
    )

    # Create Grid output with target mask (3x3 grid)
    ds_out = xe.util.cf_grid_2d(0, 10, 10.0 / 3.0, 0, 5, 5.0 / 3.0)
    target_mask_2d = np.ones((3, 3), dtype=bool)
    target_mask_2d[-1, :] = False
    ds_out['target_mask'] = xr.DataArray(target_mask_2d, dims=['lat', 'lon'])

    # Create Grid output with target mask (3x3 grid)
    ds_out = xe.util.cf_grid_2d(0, 10, 10.0 / 3.0, 0, 5, 5.0 / 3.0)
    target_mask_2d = np.ones((3, 3), dtype=bool)
    target_mask_2d[-1, :] = False
    ds_out['mask'] = xr.DataArray(target_mask_2d, dims=['lat', 'lon'])

    # Generate weights
    regridder = xe.Regridder(
        ds_in=locstream_in, ds_out=ds_out, method='nearest_s2d', locstream_in=True
    )

    # Apply weights and check results - the northmost cells should be masked
    da_out = regridder(locstream_in)['var']
    assert np.all(np.isnan(da_out[-1, :]))
    assert np.all(da_out[:-1, :] == 1)


def test_input_output_dims():
    # The only way that inputs_dims is necessary is when numpy arrays are passed through dictionaries
    grid_in = {'lon': ds_in.lon.values, 'lat': ds_in.lat.values}
    grid_out = {'lon': ds_out.lon.values, 'lat': ds_out.lat.values}

    regridder = xe.Regridder(
        grid_in, grid_out, 'bilinear', input_dims=('y', 'x'), output_dims=('y2', 'x2')
    )
    ds_result = regridder(ds_in.data.T)

    # Check the dimensions of the output
    assert 'y2' in ds_result.dims
    assert 'x2' in ds_result.dims

    # Check that the shapes match the output grid
    assert ds_result['y2'].shape == ds_out['y'].shape
    assert ds_result['x2'].shape == ds_out['x'].shape



"""
Sparse matrix multiplication (SMM) using scipy.sparse library.
"""

import warnings
from pathlib import Path

import numba as nb
import numpy as np
import sparse as sps
import xarray as xr


def read_weights(weights, n_in, n_out):
    """
    Read regridding weights into a DataArray (sparse COO matrix).

    Parameters
    ----------
    weights : str, Path, xr.Dataset, xr.DataArray, sparse.COO
        Weights generated by ESMF. Can be a path to a netCDF file generated by ESMF, an xarray.Dataset,
        a dictionary created by `ESMPy.api.Regrid.get_weights_dict` or directly the sparse
        array as returned by this function.

    n_in, n_out : integers
        ``(N_out, N_in)`` will be the shape of the returning sparse matrix.
        They are the total number of grid boxes in input and output grids::

              N_in = Nx_in * Ny_in
              N_out = Nx_out * Ny_out

        We need them because the shape cannot always be inferred from the
        largest column and row indices, due to unmapped grid boxes.

    Returns
    -------
    xr.DataArray
      A DataArray backed by a sparse.COO array, with dims ('out_dim', 'in_dim')
      and size (n_out, n_in).
    """
    if isinstance(weights, (str, Path, xr.Dataset, dict)):
        weights = _parse_coords_and_values(weights, n_in, n_out)

    elif isinstance(weights, sps.COO):
        weights = xr.DataArray(weights, dims=('out_dim', 'in_dim'), name='weights')

    elif not isinstance(weights, xr.DataArray):
        raise ValueError(f'Weights of type {type(weights)} not understood.')

    return weights


def _parse_coords_and_values(indata, n_in, n_out):
    """Creates a sparse.COO array from weights stored in a dict-like fashion.

    Parameters
    ----------
    indata: str, Path, xr.Dataset or dict
      A dictionary as returned by ESMF.Regrid.get_weights_dict
      or an xarray Dataset (or its path) as saved by xESMF.
    n_in : int
      The number of points in the input grid.
    n_out : int
      The number of points in the output grid.

    Returns
    -------
    sparse.COO
      Sparse array in the COO format.
    """

    if isinstance(indata, (str, Path, xr.Dataset)):
        if not isinstance(indata, xr.Dataset):
            if not Path(indata).exists():
                raise IOError(f'Weights file not found on disk.\n{indata}')
            ds_w = xr.open_dataset(indata)
        else:
            ds_w = indata

        if not {'col', 'row', 'S'}.issubset(ds_w.variables):
            raise ValueError(
                'Weights dataset should have variables `col`, `row` and `S` storing the indices and '
                'values of weights.'
            )

        col = ds_w['col'].values - 1  # Python starts with 0
        row = ds_w['row'].values - 1
        s = ds_w['S'].values

    elif isinstance(indata, dict):
        if not {'col_src', 'row_dst', 'weights'}.issubset(indata.keys()):
            raise ValueError(
                'Weights dictionary should have keys `col_src`, `row_dst` and `weights` storing the '
                'indices and values of weights.'
            )
        col = indata['col_src'] - 1
        row = indata['row_dst'] - 1
        s = indata['weights']

    crds = np.stack([row, col])
    return xr.DataArray(sps.COO(crds, s, (n_out, n_in)), dims=('out_dim', 'in_dim'), name='weights')


def check_shapes(indata, weights, shape_in, shape_out):
    """Compare the shapes of the input array, the weights and the regridder and raises
    potential errors.

    Parameters
    ----------
    indata : array
      Input array with the two spatial dimensions at the end,
      which should fit shape_in.
    weights : array
      Weights 2D array of shape (out_dim, in_dim).
      First element should be the product of shape_out.
      Second element should be the product of shape_in.
    shape_in : 2-tuple of int
      Shape of the input of the Regridder.
    shape_out : 2-tuple of int
      Shape of the output of the Regridder.

    Raises
    ------
    ValueError
      If any of the conditions is not respected.
    """
    # COO matrix is fast with F-ordered array but slow with C-array, so we
    # take in a C-ordered and then transpose)
    # (CSR or CRS matrix is fast with C-ordered array but slow with F-array)
    if hasattr(indata, 'flags') and not indata.flags['C_CONTIGUOUS']:
        warnings.warn('Input array is not C_CONTIGUOUS. ' 'Will affect performance.')

    # Limitation from numba : some big-endian dtypes are not supported.
    try:
        nb.from_dtype(indata.dtype)
        nb.from_dtype(weights.dtype)
    except (NotImplementedError, nb.core.errors.NumbaError):
        warnings.warn(
            'Input array has a dtype not supported by sparse and numba.'
            'Computation will fall back to scipy.'
        )

    # get input shape information
    shape_horiz = indata.shape[-2:]

    if shape_horiz != shape_in:
        raise ValueError(
            f'The horizontal shape of input data is {shape_horiz}, different from that '
            f'of the regridder {shape_in}!'
        )

    if shape_in[0] * shape_in[1] != weights.shape[1]:
        raise ValueError('ny_in * nx_in should equal to weights.shape[1]')

    if shape_out[0] * shape_out[1] != weights.shape[0]:
        raise ValueError('ny_out * nx_out should equal to weights.shape[0]')


def apply_weights(weights, indata, shape_in, shape_out):
    """
    Apply regridding weights to data.

    Parameters
    ----------
    weights : sparse COO matrix
      Regridding weights.
    indata : numpy array of shape ``(..., n_lat, n_lon)`` or ``(..., n_y, n_x)``.
        Should be C-ordered. Will be then tranposed to F-ordered.
    shape_in, shape_out : tuple of two integers
        Input/output data shape.
        For rectilinear grid, it is just ``(n_lat, n_lon)``.

    Returns
    -------
    outdata : numpy array of shape ``(..., shape_out[0], shape_out[1])``.
        Extra dimensions are the same as `indata`.
        If input data is C-ordered, output will also be C-ordered.
    """
    extra_shape = indata.shape[0:-2]

    # Limitation from numba : some big-endian dtypes are not supported.
    indata_dtype = indata.dtype
    try:
        nb.from_dtype(indata.dtype)
        nb.from_dtype(weights.dtype)
    except (NotImplementedError, nb.core.errors.NumbaError):
        indata = indata.astype('<f8')  # On the fly conversion

    # Dot product
    outdata = np.tensordot(
        indata,
        weights,
        axes=((indata.ndim - 2, indata.ndim - 1), (weights.ndim - 2, weights.ndim - 1)),
    )

    # Ensure same dtype as the input.
    outdata = outdata.astype(indata_dtype)

    # Ensure output shape is what is expected
    outdata = outdata.reshape(*extra_shape, shape_out[0], shape_out[1])
    return outdata


def post_apply_target_mask_to_weights(weights, target_mask_2d):
    """
    Set all contributions to masked target grid cells to zero.

    Parameters
    ----------
    weights : DataArray backed by a sparse.COO array
      Sparse weights matrix.
    target_mask_2d : array-like
      Mask array of shape (nx, ny) for the target grid.
      False / 0 indicates a masked cell whose weights shall be set to zero.

    Returns
    -------
    DataArray backed by a sparse.COO array
      Modified sparse weights matrix with all rows corresponding to masked target cells set to zero.

    Notes
    -----
    This defines a post-processing step applied after ESMF weight generation.
    It is useful in cases where ESMF/ESMPy masks cannot be used directly,
    which is the case when source or target are LocStream objects / sequences.
    """
    # Ensure mask can be converted to array
    try:
        target_mask_2d = np.asarray(target_mask_2d, dtype=weights.data.data.dtype)
    except Exception as e:
        raise TypeError(
            "Argument 'target_mask_2d' must be array-like and convertible to a numeric/boolean array"
        ) from e

    # Validate dimensionality and shape
    if target_mask_2d.ndim != 2:
        raise ValueError(f"Argument 'target_mask_2d' must be 2D, got shape {target_mask_2d.shape}")
    n_target, n_source = weights.data.shape
    if target_mask_2d.size != n_target:
        raise ValueError(
            f"Mismatch: weight matrix has {n_target} target cells, "
            f"but mask has {target_mask_2d.size} elements"
        )

    # Flatten mask array to align with weight matrix target index (Fortran order for ESMF layout)
    target_mask_flat = target_mask_2d.ravel(order='F')

    # Multiply row-wise by mask to zero out weights of masked target cells
    W = weights.data * target_mask_flat[:, None]

    # Create weights DataArray backed by sparse.COO
    weights = xr.DataArray(
        sps.COO(coords=W.coords, data=W.data, shape=W.shape), dims=('out_dim', 'in_dim')
    )
    return weights


def add_nans_to_weights(weights):
    """Add nan in empty rows of the regridding weights sparse matrix.

    By default, empty rows in the weights sparse matrix are interpreted as zeroes. This can become problematic
    when the field being interpreted has legitimate null values. This function inserts nan values in each row to
    make sure empty weights are propagated as nans instead of zeros. It also removes unnecessary entries, ones
    where the data is the same as the fill value (0).

    Parameters
    ----------
    weights : DataArray backed by a sparse.COO array
      Sparse weights matrix.

    Returns
    -------
    DataArray backed by a sparse.COO array
      Sparse weights matrix.
    """
    # Taken from @trondkr and adapted by @raphaeldussin to use `lil`, translated to COO by @aulemahal
    coo = weights.data
    coords = coo.coords
    data = coo.data
    # Remove unnecessary entries (roundtrip through scipy's lil did that implicitely)
    coords = coords[:, data != coo.fill_value]
    data = data[data != coo.fill_value]

    # Replace rows with no weights with a NaN at element 0, so that remapped elements are NaNs instead of zeros.
    # Fin rows with no entry in the weights, the unmapped ones
    unmapped_rows = set(np.arange(coo.shape[0])) - set(coords[0])
    # Generate one coord bper unmapped row
    new_coords = np.array([list(unmapped_rows), [0] * len(unmapped_rows)], dtype=coords.dtype)
    # Assign a NaN to the new coord so the scalar product of that row gives a NaN
    new_data = np.full((len(unmapped_rows),), np.nan)

    # Recreate the new COO weights matrix
    new = sps.COO(
        np.hstack((coords, new_coords)),
        np.hstack((data, new_data)),
        coo.shape,
        fill_value=coo.fill_value,
    )
    weights = weights.copy(data=new)
    return weights


def mask_source_indices(weights, source_indices_to_mask):
    """
    Remove entries in a sparse.COO weight matrix that map from masked source indices.

    Parameters
    ----------
    weights: DataArray backed by a sparse.COO array
      Sparse weights matrix.
    source_indices_to_mask: array-like
      Flat indices of source grid cells whose contribution should be removed
      (eg. output of xesmf.util._get_edge_indices_2d)

    Returns
    -------
    DataArray backed by a sparse.COO array:
      New weight matrix with masked source contributions removed
    """
    # Extract the sparse.COO remapping weight matrix (ntarget, nsource)
    W = weights.data

    # W.coords is a 2D array with shape (2, N), holding the non-zero entries in the
    #   sparse matrix. Rows being [0] target_idx, [1] source_idx.
    tgt_idx = W.coords[0]
    src_idx = W.coords[1]
    # Each W.data[i] represents the weight from source_idx[i] to target_idx[i]
    data = W.data

    # Validate source_indices_to_mask
    n_source = W.shape[1]
    invalid = np.asarray(source_indices_to_mask) >= n_source
    if np.any(invalid) or np.any(np.asarray(source_indices_to_mask) < 0):
        raise ValueError(
            f"Some of the provided source indices are out of valid range [0, {n_source}). "
            f"Invalid indices: {np.asarray(source_indices_to_mask)[invalid]}"
        )

    # Boolean mask for the source_idx - False for masked source indices
    mask = ~np.isin(src_idx, source_indices_to_mask)

    # Create new sparse matrix with only non-masked entries
    data_masked = data[mask]
    # Create new coordinates array by vertical (row-wise) stacking of the new target and source indices
    coords_masked = np.vstack([tgt_idx[mask], src_idx[mask]])

    # Create new sparse weight matrix and assign it to the weights DataArray
    weights = xr.DataArray(
        sps.COO(coords_masked, data_masked, shape=W.shape), dims=('out_dim', 'in_dim')
    )

    return weights


def gen_mask_from_weights(weights, nlat, nlon):
    """Generate a 2D mask from the regridding weights sparse matrix.
    This function will generate a 2D binary mask out of a regridding weights sparse matrix.

    Parameters
    ----------
    weights : DataArray backed by a sparse.COO array
      Sparse weights matrix.

    Returns
    -------
    numpy.ndarray of type numpy.int32 and of shape (nlat, nlon)
        Binary mask.
    """
    # Taken from @trondkr and adapted by @raphaeldussin to use `lil`.
    # lil matrix is better than CSR when changing sparsity
    m = weights.data.to_scipy_sparse().tolil()

    # Create mask ndarray of ones and fill with 0-elements
    mask = np.ones((nlat, nlon), dtype=np.int32).ravel()
    for krow in range(len(m.rows)):
        if any([np.isnan(x) for x in m.data[krow]]):
            mask[krow] = 0

    # Reshape and return
    return mask.reshape((nlat, nlon))


def _combine_weight_multipoly(weights, areas, indexes):
    """Reduce a weight sparse matrix (csc format) by combining (adding) columns.

    This is used to sum individual weight matrices from multi-part geometries.

    Parameters
    ----------
    weights : DataArray
      Usually backed by a sparse.COO array, with dims ('out_dim', 'in_dim')
    areas : np.array
      Array of destination areas, following same order as weights.
    indexes : array of integers
      Columns with the same "index" will be summed into a single column at this
      index in the output matrix.

    Returns
    -------
    sparse matrix (CSC)
      Sum of weights from individual geometries.
    """

    sub_weights = weights.rename(out_dim='subgeometries')

    # Create a sparse DataArray with the mesh areas
    # This ties the `out_dim` (the dimension for the original geometries) to the
    # subgeometries dimension (the exploded polygon exteriors and interiors).
    crds = np.stack([indexes, np.arange(len(indexes))])
    a = xr.DataArray(
        sps.COO(crds, areas, (indexes.max() + 1, len(indexes)), fill_value=0),
        dims=('out_dim', 'subgeometries'),
        name='area',
    )

    # Weight the regridding weights by the area of the destination polygon and sum over sub-geometries
    out = (sub_weights * a).sum(dim='subgeometries')

    # Renormalize weights along in_dim
    wsum = out.sum('in_dim')

    # Change the fill_value to 1
    wsum = wsum.copy(
        data=sps.COO(wsum.data.coords, wsum.data.data, shape=wsum.data.shape, fill_value=1)
    )

    return out / wsum



import dask
import pytest


@pytest.fixture(scope='function')
def threaded_scheduler():
    with dask.config.set(scheduler='threads'):
        yield


@pytest.fixture(scope='function')
def processes_scheduler():
    with dask.config.set(scheduler='processes'):
        yield


@pytest.fixture(scope='function')
def distributed_scheduler():
    from dask.distributed import Client, LocalCluster

    cluster = LocalCluster(threads_per_worker=1, n_workers=2, processes=True)
    client = Client(cluster)
    yield
    client.close()
    del client
    cluster.close()
    del cluster


def pytest_addoption(parser):
    parser.addoption('--runtestcases', action='store_true', default=False, help='run test cases')


def pytest_configure(config):
    config.addinivalue_line('markers', 'testcases: mark test cases')


def pytest_collection_modifyitems(config, items):
    if config.getoption('--runtestcases'):
        # --runtestcases given in cli: do not skip test cases
        return
    skip_testcases = pytest.mark.skip(reason='need --runtestcases option to run')
    for item in items:
        if 'testcases' in item.keywords:
            item.add_marker(skip_testcases)



import numpy as np
import pytest
import sparse as sps
import xarray as xr

import xesmf as xe


def test_add_nans_to_weights():
    """testing adding Nans to empty rows in sparse matrix"""
    # create input sparse matrix with one empty row (j=2)
    coords = np.array([[0, 3, 1, 0], [0, 3, 1, 2]])
    data = np.array([4.0, 5.0, 7.0, 9.0])
    Matin = sps.COO(coords, data, shape=(4, 4))

    # this is what is expected to come out (Nan added at i=0, j=2)
    coords = np.array([[0, 3, 1, 0, 2], [0, 3, 1, 2, 0]])
    data = np.array([4.0, 5.0, 7.0, 9.0, np.nan])
    expected = sps.COO(coords, data, shape=(4, 4))

    Matout = xe.smm.add_nans_to_weights(xr.DataArray(Matin, dims=('in', 'out')))
    assert np.allclose(expected.todense(), Matout.data.todense(), equal_nan=True)

    # Matrix without empty rows should return the same
    coords = np.array([[0, 3, 1, 0, 2], [0, 3, 1, 2, 1]])
    data = np.array([4.0, 5.0, 7.0, 9.0, 10.0])
    Matin = sps.COO(coords, data, shape=(4, 4))

    Matout = xe.smm.add_nans_to_weights(xr.DataArray(Matin, dims=('in', 'out')))
    assert np.allclose(Matin.todense(), Matout.data.todense())


def test_mask_source_indices():
    # Create a small sparse matrix
    coords = np.array([[0, 1, 2, 3, 4, 8], [0, 2, 4, 10, 11, 19]])
    data = np.array([0.3, 0.5, 0.2, 0.26, 0.7, 0.25])
    shape = (10, 20)
    W = sps.COO(coords, data, shape=shape)
    weights = xr.DataArray(W, dims=('out_dim', 'in_dim'))

    masked = xe.smm.mask_source_indices(weights, source_indices_to_mask=[4, 11])

    # Only selected entries should remain
    expected_coords = np.array([[0, 1, 3, 8], [0, 2, 10, 19]])
    expected_data = np.array([0.3, 0.5, 0.26, 0.25])

    assert np.array_equal(masked.data.coords, expected_coords)
    assert np.allclose(masked.data.data, expected_data)


def test_gen_mask_from_weights():
    """testing creating mask out of weight matrix Nans"""
    # Create input and output Dataset
    ds_in = xe.util.grid_2d(20, 40, 1, 20, 30, 1)
    ds_out = xe.util.grid_2d(20, 40, 2, 20, 30, 2)

    # Create random mask for ds_out
    mask = np.random.randint(low=0, high=2, size=(5, 10), dtype=np.int32)
    ds_out['mask'] = xr.DataArray(data=mask, dims=['y', 'x'])

    # Create remapping weights
    Weights = xe.Regridder(ds_in, ds_out, method='bilinear').weights

    # Generate mask from weights
    maskwgts = xe.smm.gen_mask_from_weights(Weights, 5, 10)

    # Assert equality between both masks
    assert np.array_equal(mask, maskwgts, equal_nan=False)


def test_post_apply_target_mask_to_weights():
    # Create a small sparse weights matrix with shape (9 target, 4 source)
    #  coords = [[target_indices], [source_indices]]
    coords = np.array([[0, 1, 1, 2, 3, 3, 4, 5], [0, 0, 1, 1, 2, 3, 2, 3]])
    data = np.array([0.1, 0.2, 0.3, 0.4, 0.5, 0.45, 0.7, 0.8])
    shape = (6, 4)
    W_sparse = sps.COO(coords, data, shape=shape)
    weights = xr.DataArray(W_sparse, dims=('out_dim', 'in_dim'))

    # Define a 3x3 mask for target (flattened size = 9):
    #   If all goes to plan, weights of cells 3 and 4 (i.e. index 2 and 3)
    #   will be set to 0.
    target_mask_2d = np.array([[True, False], [True, True], [False, True]])

    # Apply mask
    masked_weights = xe.smm.post_apply_target_mask_to_weights(weights, target_mask_2d)

    # Check results
    np.testing.assert_array_equal(masked_weights.data.data, np.array([0.1, 0.2, 0.3, 0.7, 0.8]))
    np.testing.assert_array_equal(
        masked_weights.data.coords, np.array([[0, 1, 1, 4, 5], [0, 0, 1, 2, 3]])
    )


def test_post_apply_target_mask_to_weights_exceptions():
    # Create a weights DataArray & mask
    coords = np.array([[0, 1], [0, 1]])
    data = np.array([0.5, 0.5])
    shape = (2, 2)
    W_sparse = sps.COO(coords, data, shape=shape)
    weights = xr.DataArray(W_sparse, dims=('out_dim', 'in_dim'))
    valid_mask = np.array([[True, False]])

    # Mask not array-like
    with pytest.raises(
        TypeError,
        match="Argument 'target_mask_2d' must be array-like and convertible to a numeric/boolean array",
    ):
        xe.smm.post_apply_target_mask_to_weights(weights, 'not_array_like')

    # Shape mismatch
    wrong_shape_mask = np.array([[True, False, True]])
    with pytest.raises(
        ValueError, match='Mismatch: weight matrix has 2 target cells, but mask has 3 elements'
    ):
        xe.smm.post_apply_target_mask_to_weights(weights, wrong_shape_mask)

    # Mask not 2D
    wrong_shape_mask = np.array([[[True]], [[True]]])
    with pytest.raises(
        ValueError, match="Argument 'target_mask_2d' must be 2D, got shape \\(2, 1, 1\\)"
    ):
        xe.smm.post_apply_target_mask_to_weights(weights, wrong_shape_mask)

    # That should work
    xe.smm.post_apply_target_mask_to_weights(weights, valid_mask)



"""
Backend for xESMF. This module wraps ESMPy's complicated API and can create
ESMF Grid and Regrid objects only using basic numpy arrays.

General idea:

1) Only use pure numpy array in this low-level backend. xarray should only be
used in higher-level APIs which interface with this low-level backend.

2) Use simple, procedural programming here. Because ESMPy Classes are
complicated enough, building new Classes will make debugging very difficult.

3) Add some basic error checking in this wrapper level.
ESMPy is hard to debug because the program often dies in the Fortran level.
So it would be helpful to catch some common mistakes in Python level.
"""

import os
import warnings
from collections.abc import Sequence

try:
    import esmpy as ESMF
except ImportError:
    import ESMF
import numpy as np
import numpy.lib.recfunctions as nprec


def warn_f_contiguous(a):
    """
    Give a warning if input array if not Fortran-ordered.

    ESMPy expects Fortran-ordered array. Passing C-ordered array will slow down
    performance due to memory rearrangement.

    Parameters
    ----------
    a : numpy array
    """
    if not a.flags['F_CONTIGUOUS']:
        warnings.warn('Input array is not F_CONTIGUOUS. ' 'Will affect performance.')


def warn_lat_range(lat):
    """
    Give a warning if latitude is outside of [-90, 90]

    Longitute, on the other hand, can be in any range,
    since the it the transform is done in (x, y, z) space.

    Parameters
    ----------
    lat : numpy array
    """
    if (lat.max() > 90.0) or (lat.min() < -90.0):
        warnings.warn('Latitude is outside of [-90, 90]')


class Grid(ESMF.Grid):
    @classmethod
    def from_xarray(cls, lon, lat, periodic=False, mask=None):
        """
        Create an ESMF.Grid object, for constructing ESMF.Field and ESMF.Regrid.

        Parameters
        ----------
        lon, lat : 2D numpy array
             Longitute/Latitude of cell centers.

             Recommend Fortran-ordering to match ESMPy internal.

             Shape should be ``(Nlon, Nlat)`` for rectilinear grid,
             or ``(Nx, Ny)`` for general quadrilateral grid.

        periodic : bool, optional
            Periodic in longitude? Default to False.
            Only useful for source grid.

        mask : 2D numpy array, optional
            Grid mask. According to the ESMF convention, masked cells
            are set to 0 and unmasked cells to 1.

            Shape should be ``(Nlon, Nlat)`` for rectilinear grid,
            or ``(Nx, Ny)`` for general quadrilateral grid.

        Returns
        -------
        grid : ESMF.Grid object
        """

        # ESMPy expects Fortran-ordered array.
        # Passing C-ordered array will slow down performance.
        for a in [lon, lat]:
            warn_f_contiguous(a)

        warn_lat_range(lat)

        # ESMF.Grid can actually take 3D array (lon, lat, radius),
        # but regridding only works for 2D array
        assert lon.ndim == 2, 'Input grid must be 2D array'
        assert lon.shape == lat.shape, 'lon and lat must have same shape'

        staggerloc = ESMF.StaggerLoc.CENTER  # actually just integer 0

        if periodic:
            num_peri_dims = 1
        else:
            num_peri_dims = None

        # ESMPy documentation claims that if staggerloc and coord_sys are None,
        # they will be set to default values (CENTER and SPH_DEG).
        # However, they actually need to be set explicitly,
        # otherwise grid._coord_sys and grid._staggerloc will still be None.
        grid = cls(
            np.array(lon.shape),
            staggerloc=staggerloc,
            coord_sys=ESMF.CoordSys.SPH_DEG,
            num_peri_dims=num_peri_dims,
        )

        # The grid object points to the underlying Fortran arrays in ESMF.
        # To modify lat/lon coordinates, need to get pointers to them
        lon_pointer = grid.get_coords(coord_dim=0, staggerloc=staggerloc)
        lat_pointer = grid.get_coords(coord_dim=1, staggerloc=staggerloc)

        # Use [...] to avoid overwritting the object. Only change array values.
        lon_pointer[...] = lon
        lat_pointer[...] = lat

        # Follows SCRIP convention where 1 is unmasked and 0 is masked.
        # See https://github.com/NCPP/ocgis/blob/61d88c60e9070215f28c1317221c2e074f8fb145/src/ocgis/regrid/base.py#L391-L404
        if mask is not None:
            # remove fractional values
            mask = np.where(mask == 0, 0, 1)
            # convert array type to integer (ESMF compat)
            grid_mask = mask.astype(np.int32)
            if not (grid_mask.shape == lon.shape):
                raise ValueError(
                    'mask must have the same shape as the latitude/longitude '
                    'coordinates, got: mask.shape = %s, lon.shape = %s' % (mask.shape, lon.shape)
                )
            grid.add_item(ESMF.GridItem.MASK, staggerloc=ESMF.StaggerLoc.CENTER, from_file=False)
            grid.mask[0][:] = grid_mask

        return grid

    def get_shape(self, loc=ESMF.StaggerLoc.CENTER):
        """Return shape of grid for specified StaggerLoc"""
        # We cast explicitly to python's int (numpy >=2)
        return tuple(map(int, self.size[loc]))


class LocStream(ESMF.LocStream):
    @classmethod
    def from_xarray(cls, lon, lat):
        """
        Create an ESMF.LocStream object, for contrusting ESMF.Field and ESMF.Regrid

        Parameters
        ----------
        lon, lat : 1D numpy array
             Longitute/Latitude of cell centers.

        Returns
        -------
        locstream : ESMF.LocStream object
        """

        if len(lon.shape) > 1:
            raise ValueError('lon can only be 1d')
        if len(lat.shape) > 1:
            raise ValueError('lat can only be 1d')

        assert lon.shape == lat.shape

        location_count = len(lon)

        locstream = cls(location_count, coord_sys=ESMF.CoordSys.SPH_DEG)

        locstream['ESMF:Lon'] = lon.astype(np.dtype('f8'))
        locstream['ESMF:Lat'] = lat.astype(np.dtype('f8'))

        return locstream

    def get_shape(self):
        """Return LocStream shape."""
        return (self.size, 1)


def add_corner(grid, lon_b, lat_b):
    """
    Add corner information to ESMF.Grid for conservative regridding.

    Not needed for other methods like bilinear or nearest neighbour.

    Parameters
    ----------
    grid : ESMF.Grid object
        Generated by ``Grid.from_xarray()``. Will be modified in-place.

    lon_b, lat_b : 2D numpy array
        Longitute/Latitude of cell corner
        Recommend Fortran-ordering to match ESMPy internal.
        Shape should be ``(Nlon+1, Nlat+1)``, or ``(Nx+1, Ny+1)``
    """

    # codes here are almost the same as Grid.from_xarray(),
    # except for the "staggerloc" keyword
    staggerloc = ESMF.StaggerLoc.CORNER  # actually just integer 3

    for a in [lon_b, lat_b]:
        warn_f_contiguous(a)

    warn_lat_range(lat_b)

    assert lon_b.ndim == 2, 'Input grid must be 2D array'
    assert lon_b.shape == lat_b.shape, 'lon_b and lat_b must have same shape'
    assert np.array_equal(lon_b.shape, grid.max_index + 1), 'lon_b should be size (Nx+1, Ny+1)'
    assert (grid.num_peri_dims == 0) and (
        grid.periodic_dim is None
    ), 'Cannot add corner for periodic grid'

    grid.add_coords(staggerloc=staggerloc)

    lon_b_pointer = grid.get_coords(coord_dim=0, staggerloc=staggerloc)
    lat_b_pointer = grid.get_coords(coord_dim=1, staggerloc=staggerloc)

    lon_b_pointer[...] = lon_b
    lat_b_pointer[...] = lat_b


class Mesh(ESMF.Mesh):
    @classmethod
    def from_polygons(cls, polys, element_coords='centroid'):
        """
        Create an ESMF.Mesh object from a list of polygons.

        All exterior ring points are added to the mesh as nodes and each polygon
        is added as an element, with the polygon centroid as the element's coordinates.

        Parameters
        ----------
        polys : sequence of shapely Polygon
           Holes are not represented by the Mesh.
        element_coords : array or "centroid", optional
            If "centroid", the polygon centroids will be used (default)
            If an array of shape (len(polys), 2) : the element coordinates of the mesh.
            If None, the Mesh's elements will not have coordinates.

        Returns
        -------
        mesh : ESMF.Mesh
            A mesh where each polygon is represented as an Element.
        """
        node_num = sum(len(e.exterior.coords) - 1 for e in polys)
        elem_num = len(polys)
        # Pre alloc arrays. Special structure for coords makes the code faster.
        crd_dt = np.dtype([('x', np.float32), ('y', np.float32)])
        node_coords = np.empty(node_num, dtype=crd_dt)
        node_coords[:] = (np.nan, np.nan)  # Fill with impossible values
        element_types = np.empty(elem_num, dtype=np.uint32)
        element_conn = np.empty(node_num, dtype=np.uint32)
        # Flag for centroid calculation
        calc_centroid = isinstance(element_coords, str) and element_coords == 'centroid'
        if calc_centroid:
            element_coords = np.empty(elem_num, dtype=crd_dt)
        inode = 0
        iconn = 0
        for ipoly, poly in enumerate(polys):
            ring = poly.exterior
            if calc_centroid:
                element_coords[ipoly] = poly.centroid.coords[0]
            element_types[ipoly] = len(ring.coords) - 1
            for coord in ring.coords[:-1] if ring.is_ccw else ring.coords[:0:-1]:
                crd = np.asarray(coord, dtype=crd_dt)  # Cast so we can compare
                node_index = np.where(node_coords == crd)[0]
                if node_index.size == 0:  # New node
                    node_coords[inode] = crd
                    element_conn[iconn] = inode
                    inode += 1
                else:  # Node already exists
                    element_conn[iconn] = node_index[0]
                iconn += 1
        node_num = inode  # With duplicate nodes, inode < node_num

        mesh = cls(2, 2, coord_sys=ESMF.CoordSys.SPH_DEG)
        mesh.add_nodes(
            node_num,
            np.arange(node_num) + 1,
            nprec.structured_to_unstructured(node_coords[:node_num]).ravel(),
            np.zeros(node_num),
        )
        if calc_centroid:
            element_coords = nprec.structured_to_unstructured(element_coords)
        if element_coords is not None:
            element_coords = element_coords.ravel()
        try:
            mesh.add_elements(
                elem_num,
                np.arange(elem_num) + 1,
                element_types,
                element_conn,
                element_coords=element_coords,
            )
        except ValueError as err:
            raise ValueError(
                'ESMF failed to create the Mesh, this usually happen when some polygons are invalid (test with `poly.is_valid`)'
            ) from err
        return mesh

    def get_shape(self, loc=ESMF.MeshLoc.ELEMENT):
        """Return the shape of the Mesh at specified MeshLoc location."""
        return (self.size[loc], 1)


def esmf_regrid_build(  # noqa: C901
    sourcegrid,
    destgrid,
    method,
    filename=None,
    extra_dims=None,
    extrap_method=None,
    extrap_dist_exponent=None,
    extrap_num_src_pnts=None,
    ignore_degenerate=None,
    vector_regrid=None,
):
    """
    Create an ESMF.Regrid object, containing regridding weights.

    Parameters
    ----------
    sourcegrid, destgrid : ESMF.Grid or ESMF.Mesh object
        Source and destination grids.

        Should create them by ``Grid.from_xarray()``
        (with optionally ``add_corner()``),
        instead of ESMPy's original API.

    method : str
        Regridding method. Options are

        - 'bilinear'
        - 'conservative', **need grid corner information**
        - 'conservative_normed', **need grid corner information**
        - 'patch'
        - 'nearest_s2d'
        - 'nearest_d2s'

    filename : str, optional
        Offline weight file. **Require ESMPy 7.1.0.dev38 or newer.**
        With the weights available, we can use Scipy's sparse matrix
        multiplication to apply weights, which is faster and more Pythonic
        than ESMPy's online regridding. If None, weights are stored in
        memory only.

    extra_dims : a list of integers, optional
        Extra dimensions (e.g. time or levels) in the data field

        This does NOT affect offline weight file, only affects online regrid.

        Extra dimensions will be stacked to the fastest-changing dimensions,
        i.e. following Fortran-like instead of C-like conventions.
        For example, if extra_dims=[Nlev, Ntime], then the data field dimension
        will be [Nlon, Nlat, Nlev, Ntime]

    extrap_method : str, optional
        Extrapolation method. Options are

        - 'inverse_dist'
        - 'nearest_s2d'

    extrap_dist_exponent : float, optional
        The exponent to raise the distance to when calculating weights for the
        extrapolation method. If none are specified, defaults to 2.0

    extrap_num_src_pnts : int, optional
        The number of source points to use for the extrapolation methods
        that use more than one source point. If none are specified, defaults to 8

    ignore_degenerate : bool, optional
        If False (default), raise error if grids contain degenerated cells
        (i.e. triangles or lines, instead of quadrilaterals)

    vector_regrid : bool, optional
        If True, treat a single extra (non-spatial) dimension in the source and
        destination data fields as the components of a vector. (If True and
        there is more than one extra dimension in either the source or
        destination data fields, an error will be raised.) If not specified,
        defaults to False.

        Only vector dimensions of size 2 are supported. The first entry is
        interpreted as the east component and the second as the north component.
        i.e., ``extra_dims`` must be ``[2]``.

        Requires ESMPy 8.9.0 or newer.

    Returns
    -------
    regrid : ESMF.Regrid object

    """

    # use shorter, clearer names for options in ESMF.RegridMethod
    method_dict = {
        'bilinear': ESMF.RegridMethod.BILINEAR,
        'conservative': ESMF.RegridMethod.CONSERVE,
        'conservative_normed': ESMF.RegridMethod.CONSERVE,
        'patch': ESMF.RegridMethod.PATCH,
        'nearest_s2d': ESMF.RegridMethod.NEAREST_STOD,
        'nearest_d2s': ESMF.RegridMethod.NEAREST_DTOS,
    }
    try:
        esmf_regrid_method = method_dict[method]
    except Exception:
        raise ValueError('method should be chosen from ' '{}'.format(list(method_dict.keys())))

    # use shorter, clearer names for options in ESMF.ExtrapMethod
    extrap_dict = {
        'inverse_dist': ESMF.ExtrapMethod.NEAREST_IDAVG,
        'nearest_s2d': ESMF.ExtrapMethod.NEAREST_STOD,
        None: None,
    }
    try:
        esmf_extrap_method = extrap_dict[extrap_method]
    except KeyError:
        raise KeyError(
            '`extrap_method` should be chosen from ' '{}'.format(list(extrap_dict.keys()))
        )

    # until ESMPy updates ESMP_FieldRegridStoreFile, extrapolation is not possible
    # if files are written on disk
    if (extrap_method is not None) & (filename is not None):
        raise ValueError('`extrap_method` cannot be used along with `filename`.')

    # conservative regridding needs cell corner information
    if method in ['conservative', 'conservative_normed']:
        if not isinstance(sourcegrid, ESMF.Mesh) and not sourcegrid.has_corners:
            raise ValueError(
                'source grid has no corner information. ' 'cannot use conservative regridding.'
            )
        if not isinstance(destgrid, ESMF.Mesh) and not destgrid.has_corners:
            raise ValueError(
                'destination grid has no corner information. ' 'cannot use conservative regridding.'
            )

    if vector_regrid:
        # Check this ESMPy requirement in order to give a more helpful error message if it
        # isn't met
        if not (isinstance(extra_dims, Sequence) and len(extra_dims) == 1 and extra_dims[0] == 2):
            raise ValueError('`vector_regrid` currently requires `extra_dims` to be `[2]`')

    # ESMF.Regrid requires Field (Grid+data) as input, not just Grid.
    # Extra dimensions are specified when constructing the Field objects,
    # not when constructing the Regrid object later on.
    if isinstance(sourcegrid, ESMF.Mesh):
        sourcefield = ESMF.Field(sourcegrid, meshloc=ESMF.MeshLoc.ELEMENT, ndbounds=extra_dims)
    else:
        sourcefield = ESMF.Field(sourcegrid, ndbounds=extra_dims)
    if isinstance(destgrid, ESMF.Mesh):
        destfield = ESMF.Field(destgrid, meshloc=ESMF.MeshLoc.ELEMENT, ndbounds=extra_dims)
    else:
        destfield = ESMF.Field(destgrid, ndbounds=extra_dims)

    # ESMF bug? when using locstream objects, options src_mask_values
    # and dst_mask_values produce runtime errors
    allow_masked_values = True
    if isinstance(sourcefield.grid, ESMF.api.locstream.LocStream):
        allow_masked_values = False
    if isinstance(destfield.grid, ESMF.api.locstream.LocStream):
        allow_masked_values = False

    # ESMPy will throw an incomprehensive error if the weight file
    # already exists. Better to catch it here!
    if filename is not None:
        assert not os.path.exists(
            filename
        ), 'Weight file already exists! Please remove it or use a new name.'

    # re-normalize conservative regridding results
    # https://github.com/JiaweiZhuang/xESMF/issues/17
    if method == 'conservative_normed':
        norm_type = ESMF.NormType.FRACAREA
    else:
        norm_type = ESMF.NormType.DSTAREA

    # Calculate regridding weights.
    # Must set unmapped_action to IGNORE, otherwise the function will fail,
    # if the destination grid is larger than the source grid.
    kwargs = dict(
        filename=filename,
        regrid_method=esmf_regrid_method,
        unmapped_action=ESMF.UnmappedAction.IGNORE,
        ignore_degenerate=ignore_degenerate,
        norm_type=norm_type,
        extrap_method=esmf_extrap_method,
        extrap_dist_exponent=extrap_dist_exponent,
        extrap_num_src_pnts=extrap_num_src_pnts,
        factors=filename is None,
    )
    if allow_masked_values:
        kwargs.update(dict(src_mask_values=[0], dst_mask_values=[0]))
    # Only add the vector_regrid argument if it is given and true; this supports backwards
    # compatibility with versions of ESMPy prior to 8.9.0 that do not have this option.
    # (If the user explicitly sets the vector_regrid argument, however, that will still be
    # passed through; this will lead to an exception with older ESMPy versions.)
    if vector_regrid:
        kwargs['vector_regrid'] = vector_regrid

    regrid = ESMF.Regrid(sourcefield, destfield, **kwargs)

    return regrid


def esmf_regrid_apply(regrid, indata):
    """
    Apply existing regridding weights to the data field,
    using ESMPy's built-in functionality.

    xESMF use Scipy to apply weights instead of this.
    This is only for benchmarking Scipy's result and performance.

    Parameters
    ----------
    regrid : ESMF.Regrid object
        Contains the mapping from the source grid to the destination grid.

        Users should create them by esmf_regrid_build(),
        instead of ESMPy's original API.

    indata : numpy array of shape ``(Nlon, Nlat, N1, N2, ...)``
        Extra dimensions ``(N1, N2, ...)`` are specified in
        ``esmf_regrid_build()``.

        Recommend Fortran-ordering to match ESMPy internal.

    Returns
    -------
    outdata : numpy array of shape ``(Nlon_out, Nlat_out, N1, N2, ...)``

    """

    # Passing C-ordered input data will be terribly slow,
    # since indata is often quite large and re-ordering memory is expensive.
    warn_f_contiguous(indata)

    # Get the pointers to source and destination fields.
    # Because the regrid object points to its underlying field&grid,
    # we can just pass regrid from ESMF_regrid_build() to ESMF_regrid_apply(),
    # without having to pass all the field&grid objects.
    sourcefield = regrid.srcfield
    destfield = regrid.dstfield

    # pass numpy array to the underlying Fortran array
    sourcefield.data[...] = indata

    # apply regridding weights
    destfield = regrid(sourcefield, destfield)

    return destfield.data


def esmf_regrid_finalize(regrid):
    """
    Free the underlying Fortran array to avoid memory leak.

    After calling ``destroy()`` on regrid or its fields, we cannot use the
    regrid method anymore, but the input and output data still exist.

    Parameters
    ----------
    regrid : ESMF.Regrid object

    """
    # We do not destroy the Grids here, as they might be reused between multiple regrids
    regrid.destroy()
    regrid.srcfield.destroy()
    regrid.dstfield.destroy()

    # double check
    assert regrid.finalized
    assert regrid.srcfield.finalized
    assert regrid.dstfield.finalized


# Deprecated as of version 0.5.0


def esmf_locstream(lon, lat):
    warnings.warn(
        '`esmf_locstream` is being deprecated in favor of `LocStream.from_xarray`',
        DeprecationWarning,
    )
    return LocStream.from_xarray(lon, lat)


def esmf_grid(lon, lat, periodic=False, mask=None):
    warnings.warn(
        '`esmf_grid` is being deprecated in favor of `Grid.from_xarray`', DeprecationWarning
    )
    return Grid.from_xarray(lon, lat)
